#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 11 08:07:21 2021
Caractérisation de la variabilité sur des essais expérimentaux
@author: xavierjourdain
"""

# Bibliothèque de calcul scientifique
import numpy as np
# Bibliothèque pour réaliser des graphiques
import matplotlib.pyplot as plt
# Visualisation de graphiques 3D
from mpl_toolkits.mplot3d import Axes3D
# Bibliothèque d'outils statistiques
from scipy import stats
# Minimisation
from scipy import optimize
import pandas as pd
import plotly.graph_objects as go

#### FONCTIONS ###############################################
# loi normale
def loi_de_probabilite(x_tmp,moyenne_tmp,ecart_type_tmp):
    #log_normale_eq = 1/(x_tmp * ecart_type_tmp * np.sqrt(2*np.pi)) * np.exp(-(np.log(x_tmp)-moyenne_tmp)**2 / (2*ecart_type_tmp**2))
    # fonction scipy : stats.lognorm.pdf(x , s, loc, scale)  <=> lognorm.pdf(y, s) / scale with y = (x - loc) / scale
    log_normale_scipy = stats.lognorm.pdf(x_tmp,s=ecart_type_tmp,loc=0,scale=np.exp(moyenne_tmp))
    return log_normale_scipy

# fonction coût pour la minimisation
def fonction_cout(vec2optimize , x_tmp , y_donnees_exp):
    valeurs_y_loi_choisie = loi_de_probabilite(x_tmp,vec2optimize[0],vec2optimize[1])
    cout = np.linalg.norm(valeurs_y_loi_choisie - y_donnees_exp)
    return cout

def moy_etype_lognorm(esperance_tmp,ecart_type_tmp):
    ecart_type_lognorm_tmp = np.sqrt(np.log(1+ecart_type_tmp**2/esperance_tmp**2))
    moyenne_lognorm_tmp = np.log(esperance_tmp) - 0.5 * ecart_type_lognorm_tmp**2
    return moyenne_lognorm_tmp,ecart_type_lognorm_tmp

##### DONNÉES À DÉFINIR PAR L'UTILISATEUR ####################

# Import du fichier de données
nom_fichier = 'Resultats_Exp_Traction_Fmax'
donnees_exp = np.genfromtxt(''+nom_fichier+'.csv', delimiter=',')
titre_donnees_exp = "Effort max [N]"
# Nombre d'intervalles réguliers pour le découpage de l'histogramme
nb_intervalles_histo = 20

#### ANALYSE DES DONNÉES #####################################
# Caractérisation des résultats expérimentaux
esperance= np.mean(donnees_exp)
ecart_type = np.std(donnees_exp,ddof=1)
moyenne_lognorm,ecart_type_lognorm =  moy_etype_lognorm(esperance,ecart_type)
minimum = np.min(donnees_exp)
maximum = np.max(donnees_exp)
nb_donnees = len(donnees_exp)
largeur_intervalles_histo = (maximum-minimum)/nb_intervalles_histo


x_interv_histo = np.linspace(minimum,maximum,nb_intervalles_histo)


# Histogramme des résultats expérimentaux
fig1 = plt.figure()
plt.title("Répartition des {} résultats expérimentaux\n(en nombre)".format(nb_donnees))
fig1.patch.set_alpha(0.0)
plt.hist(donnees_exp,bins=nb_intervalles_histo,histtype='stepfilled', label="données expérimentales")
plt.xlabel(titre_donnees_exp) 
plt.ylabel('Nombre de résultats par intervalle') 
plt.legend()
plt.show()

fig1 = plt.figure()
plt.title("Répartition des {} résultats expérimentaux\n(intégrale à 1)".format(nb_donnees))
fig1.patch.set_alpha(0.0)
[valeurs_y_donnees_exp,lim_bins,rectangles] = plt.hist(donnees_exp,bins=nb_intervalles_histo,histtype='stepfilled',density=True,label="données expérimentales")
plt.xlabel(titre_donnees_exp) 
plt.ylabel("probabilité d'occurence/\n((max-min)/nb_intervalles)") 
plt.legend()
plt.show()

#### IDNTIFICATION DE LA MOYENNE ET DE L'ÉCART-TYPE #########

#### Visualisation de la fonction coût
nb_de_pas = 200
# Intervalle parcouru par la moyenne
x1 = np.linspace(minimum,maximum,nb_de_pas)
# Intervalle parcouru par l'écart-type
x2 = np.linspace(100,800,nb_de_pas)
fonction_3D = np.zeros((nb_de_pas,nb_de_pas))
for x1_i in range(0,nb_de_pas):
    for x2_i in range(0,nb_de_pas):
        moyenne_lognorm_i,ecart_type_lognorm_i  = moy_etype_lognorm(x1[x1_i],x2[x2_i])
        fonction_3D[x2_i,x1_i] = fonction_cout([moyenne_lognorm_i ,ecart_type_lognorm_i], x_interv_histo,valeurs_y_donnees_exp)
mini = np.min(fonction_3D)
x1V,x2V = np.meshgrid(x1, x2)
fig2 = plt.figure()
ax1 = fig2.add_subplot(111, projection='3d')
ax1.set_zlabel('Valeur de la fonction coût')
mycmap = plt.get_cmap('gist_earth')
ax1.set_title("Fonction coût à minimiser\n(min = {})".format(mini))
ax1.set_xlabel('moyenne')
ax1.set_ylabel('écart-type')
surface1 = ax1.plot_surface(x1V, x2V, fonction_3D, cmap=mycmap)
fig1.colorbar(surface1, ax=ax1)
plt.show()

# Autre visualisation dans un navigateur
import plotly.io as pio
pio.renderers.default='browser'
fig = go.Figure(data=[go.Surface(z=fonction_3D,x=x1,y=x2)])
fig.show()


### identification avec la moyenne et l'écart-type des valeurs expérimentales
#vecteur initial pour l'identification
vecteur_initial = np.array([moyenne_lognorm,ecart_type_lognorm])
print(vecteur_initial)
#limites pour la moyenne et l'écart-type
limites=[(np.log(minimum),np.log(maximum)),(0,abs(np.log(maximum)-np.log(minimum)))]
print(limites)
print([(minimum,maximum),(0,abs(maximum-minimum))])
# option pour visualiser la convergence
options = dict(disp = True,gtol = 1e-10)
# 
vec_opt = optimize.minimize(fonction_cout,
                            vecteur_initial,
                            args=(x_interv_histo,valeurs_y_donnees_exp),
                            method = 'CG',
#                            bounds = limites,
                            options=options,
                            )
print("Valeurs initiales :")
esperance0 = np.exp(vecteur_initial[0]+(vecteur_initial[1]**2)*0.5)
ecart_type0 = np.sqrt(np.exp(vecteur_initial[1]**2)-1)*esperance0
print(" =>   Moyenne : {} ;  Écart-type : {}".format(esperance0,ecart_type0))
print("Après optimisation :")
esperance_opt = np.exp(vec_opt['x'][0]+(vec_opt['x'][1]**2)*0.5)
ecart_type_opt = np.sqrt(np.exp(vec_opt['x'][1]**2)-1)*esperance_opt
print(" =>   Moyenne : {} ;  Écart-type : {}".format(esperance_opt,ecart_type_opt))
moyenne_opt = vec_opt['x'][0]
ecart_type_opt = vec_opt['x'][1]
erreur_relative = fonction_cout([moyenne_opt,ecart_type_opt], x_interv_histo,valeurs_y_donnees_exp)

#### RÉCUPÉRATION VALEUR FONCTION QUANTILE #################

pourcentage = 0.05
valeur_cible = round(stats.lognorm.ppf(pourcentage,s=ecart_type_lognorm,loc=0,scale=np.exp(moyenne_lognorm)),1)

valeurs_x = np.linspace(minimum,maximum,100)
valeurs_y_loi_proba = loi_de_probabilite(valeurs_x,vec_opt['x'][0],vec_opt['x'][1])

# Graphique non normé (intégrale = nb d'essais)
fig1 = plt.figure()
plt.title("Répartition des {} résultats expérimentaux\n(en nombre)\n[erreur relative : {}%]".format(nb_donnees,round(100*erreur_relative,2)))
fig1.patch.set_alpha(0.0)
plt.plot(valeurs_x , valeurs_y_loi_proba * nb_donnees * (maximum-minimum)/nb_intervalles_histo,'r--', label='loi identifiée')
plt.hist(donnees_exp,bins=nb_intervalles_histo,histtype='stepfilled', label="données expérimentales")
plt.axvline(x=valeur_cible,color='green')
plt.xlabel(titre_donnees_exp) 
plt.ylabel('Nombre de résultats par intervalle') 
plt.legend()
plt.show()

# Graphique normé (intégrale = 1)
fig1 = plt.figure()
plt.title("Répartition des {} résultats expérimentaux\n(intégrale à 1)\n[erreur relative : {}%]".format(nb_donnees,round(100*erreur_relative,2)))
fig1.patch.set_alpha(0.0)
plt.plot(valeurs_x, valeurs_y_loi_proba,'r--', label='loi identifiée')
plt.hist(donnees_exp,bins=nb_intervalles_histo,histtype='stepfilled',density=True, label="données expérimentales")
plt.axvline(x=valeur_cible,color='green',label="quantile {}% : {}".format(pourcentage*100,valeur_cible))
plt.xlabel(titre_donnees_exp) 
plt.ylabel("probabilité d'occurence/\n((max-min)/nb_intervalles)") 
plt.legend()
plt.show()

# Sortie des valeurs souhaitées
print("Étude sur {} valeurs expérimentales :".format(nb_donnees))
print("{}% des valeurs de la loi identifiée sont au dessus de {} (unité des valeurs du fichier d'entrée)".format(100*(1-pourcentage),valeur_cible))


